/***********
 * Modules0.c
 * implementation of the Oberon module loader for the IBM RISC System/6000
 * Marc-Michael Brandis, 26.4.91 / 13.10.93
 *
 * Guenter Feldmann, modified for Solaris_PPC,    4.1.97
 *		     new Module structure	 20.2.98
 *		     new LinkEntry structure	 20.10.98
 *
 *		     MacX version		28.01.02
 *		     adapted to new compiler	24.04.02
 *                            
 *
 *
 ***********/


#include <assert.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <errno.h>
#include "Modules0.h"
#include "POP2limits.h"
#include "OberonSyntax.h"

#define bigendian	TRUE /* SolarisPPP: FALSE, LinuxPPC: TRUE */

/* block tags */
#define EntryMark 0x82
#define CommandMark 0x83
#define PointerMark 0x84
#define ImportMark 0x85
#define LinkMark 0x86
#define DataMark 0x87
#define ExportMark 0x88
#define CodeMark 0x89
#define UseMark 0x8A
#define TypeMark 0x8B
#define TrapMark 0x8C
#define RefMark 0x8D

/* link entry tags */
#define SYSMTag 0xFF
#define LinkMTag 0xFE
#define CaseETag 0xFE
#define NewRecETag 0xFF
#define NewSysETag 0xFE
#define NewArrETag 0xFD

#define EUEnd 0
#define EURecord 1
#define EUobjScope 0
#define EUrecScope 1
#define EUerrScope -1

#define EUProcFlag 0x80000000
#define EUTypFlag 0xC0000000

#define DefMaxImport 64
#define DefMaxStruct 32
#define DefMaxReimp 32
        
/* some opcodes */
#define iCAL 0x38000000
#define iCAU 0x3C000000
#define iBL 0x48000001
#define iBLmask 0x03fffffc

#define BLSIZE 4096
#define uint	unsigned long

/* types */

typedef struct _objfile {
	int   handle;
	char  name[ModuleNameLength+4];
	int   pos;
} FileDesc;

typedef FileDesc* File;

/* typedef FILE* File; */

typedef struct _link_entry {
    INTEGER mod, ent;
    INTEGER offset;
} LinkEntry;

typedef struct _import_desc {
    ModuleName name;
    LONGINT key;
} ImportDesc;


static int	u_argc;
static char**	u_argv;
static int	o_debug;

int	suid_root;

LONGINT    oberonTrapHandler;
LONGINT    oberonXErrorHandler;
LONGINT    oberonXIOErrorHandler;

static  LONGINT KernelSB;
static  CHAR* OberonPath;

static	LONGINT	newSys;
static	LONGINT newRec;
static	LONGINT newArr;

static  INTEGER res;

static	Module modList;
static	LONGINT	exportDT;

static	BOOLEAN trace;


#define tagSize 4

/* macro to align memory addresses */
#define Align8(x) (((x)+7)&0xfffffff8)

/* macros to access code words */
#define disp16(x) ((x<<16)>>16)
#define instr16(x) (x&0xffff0000)
#define split(x,u,l) {u=((x>>16)+((x>>15)&1))&0x0000ffff;l=(x&0x0000ffff);}
#define opcode(x) (x&0xfc000000)
#define disp26(x) ((x<<6)>>8)
#define instr26(x) (x&~iBLmask)

/* macros to access parts of words */
#define UHIGH(x) (((x)>>16)&0xffff)
#define ULOW(x) ((x)&0xffff)
#define SHIGH(x) ((x)>>16)
#define SLOW(x) (((x)<<16)>>16)

extern void CallOberon( LONGINT pc, LONGINT sb );

static Module ThisMod( CHAR * ); 

#include "dlsym.ci"

/* file access routines */

PROCEDURE invalidObj( File f ) {
    res = invalidObjFile;
    printf( "%s at pos %7o: invalid object format\n", f->name, f->pos );
}

static CHAR next( File f ) {
    CHAR c;

    read( f->handle, &c, 1 );
    f->pos++;
    IF trace THEN 
	printf("<%o>", c);
    END
    return c;
}

PROCEDURE Check( File f, CHAR tag ) {
    CHAR v;
    
    /* printf("tag %x\n", tag & 0xff); */
    IF (v = next( f )) != tag THEN
	invalidObj( f );
    	printf("\t\texp: %2x, got: %2x\n", tag, v );
    END
}

PROCEDURE ReadInt( File f, INTEGER* val ) {
    CHAR *p;
    CHAR c[2];

    IF read( f->handle, c, 2 ) != 2 THEN
        invalidObj( f );
    END
    p = (CHAR*)val;
    IF bigendian THEN
	*p++ = c[1]; *p = c[0];
    ELSE
	*p++ = c[0]; *p = c[1];
    END
    f->pos = f->pos + 2;
}

PROCEDURE ReadLInt( File f, LONGINT* val ) {
    CHAR *p;
    CHAR c[4];

    IF read( f->handle, c, 4 ) != 4 THEN
        invalidObj( f );
    END
    p = (CHAR*)val;
    IF bigendian THEN
	*p++ = c[3]; *p++ = c[2]; *p++ = c[1]; *p = c[0];
    ELSE
	*p++ = c[0]; *p++ = c[1]; *p++ = c[2]; *p = c[3];
    END
    f->pos = f->pos + 4;
}

PROCEDURE ReadNum( File f, LONGINT *val ) {
    CHAR c;
    LONGINT n, x;
    INTEGER s, d;
    
    IF trace THEN printf( "[" ); END
    s = 0; n = 0; 
    c = next( f );
    x = c & 0177;
    WHILE x != c DO
        n = n + (x << s); s = s + 7; 
	c = next( f );
    	x = c & 0177;
    END
    d = 25 - s;
    IF d >= 0 THEN
        n = n + (((LONGINT)(x) << 25) >> d);
    ELSE
        n = n + (((LONGINT)(x) << 25) << -d);
    END
    IF trace THEN printf( "]" ); END
    *val = n;
}


PROCEDURE ReadBlock( File f, void* ptr, LONGINT nr ) {
    IF read( f->handle, ptr, nr ) != nr THEN
	invalidObj( f );
    END
    f->pos = f->pos + nr;
}

PROCEDURE Skip( File f, INTEGER n ) {
    INTEGER l = n;
    CHAR c;
    
    WHILE l > 0  DO
        c = next( f ); l--;
    END
}


PROCEDURE ReadString( File ObjFile, CHAR* p ) {
    BOOLEAN cont;
    CHAR c;

    cont = FALSE;
    do {
	c = next( ObjFile );
	IF c == 0x7F THEN
	    c = 0; cont = TRUE;
	ELSIF c & 0x80 THEN
	    *p++ = (c - 0x80); *p = 0; return;
	ELSE
	    *p++ = c;
	END
    } while ( c != 0 );
    IF cont THEN
	do {
	    c = next( ObjFile );
	    *p++ = c; 
	} while (c != 0);
    END
}

static BOOLEAN FindFile( File f ) {
    CHAR fname[256];
    CHAR *p, *q, *r;

    strcpy( fname, f->name );
    f->pos = 0;
    IF (f->handle = open( fname, O_RDONLY )) > 0 THEN  return TRUE;  END
    p = OberonPath;
    WHILE *p != '\0' DO
        q = fname;
        WHILE *p > ' ' DO *q++ = *p++; END
        IF *(q-1) != '/' THEN  *q++ = '/'; END
        r = f->name;
        WHILE *q++ = *r++ DO END
        IF (f->handle = open( fname, O_RDONLY )) > 0 THEN  return TRUE; END
        WHILE *p > '\0' && *p <= ' ' DO  p++; END
    END
    return FALSE;
}


static Module FindModule( CHAR* name ) {
    Module m;

    m = (Module)modList;
    WHILE m != NIL && strcmp(m->name, name) != 0 DO m = m->link; END
    return m;
}


/* creation of data structures */

static Tag AllocTypeDesc( int recsize, int nofmth, int nofptrs ) {
    Tag tag;
    SuperTag supertag;
    LONGINT negpart, size, i;

    negpart = nofmth*sizeof( PROC ) + (MaxExts+1)*tagSize + tagSize;
    size = 4 + sizeof( SuperDesc ) + negpart + sizeof( TypeDesc ) + nofptrs*4;    

    supertag = (SuperTag)malloc( size );
    supertag = (SuperTag)((long)supertag + 4);
    TAG(*supertag) = ((long)supertag | 1);
    tag = (Tag)((char*)supertag + sizeof( SuperDesc ) + negpart);
    supertag->tdsize = sizeof( TypeDesc ) + nofptrs*4 + negpart;
    supertag->ptroff = -4;
    supertag->tag = tag;
    supertag->pad[0] = nofmth;
    TAG(*tag) = ((long)supertag | 1);
    tag->recsize = recsize;
    PTROFF(*tag, nofptrs) = -4;
    for (i = 0; i < nofmth; i++) METHOD(*tag, i).PC = -1;
    return tag;
}


PROCEDURE AllocateModule( Module m ) {
    LONGINT code, entries, commands, imports, typedescs, pointers, traps, refs;
    
    code = Align8( m->conSize ) + Align8( m->dataSize );
    entries = Align8( code + m->codeSize*4 );
    commands = Align8( entries + m->nofEntries*2 );
    imports = Align8( commands + m->nofCmds*sizeof( CommandDesc ) );
    typedescs = Align8( imports + (m->nofImps+1)*sizeof( Module ) );
    pointers = Align8( typedescs + m->nofTds*4 );
    traps = Align8( pointers + m->nofPtrs*4 );
    refs = Align8( traps + (m->nofTraps+1)*sizeof( TrapDesc ) );
    m->blockSize = Align8( refs + m->refSize + 8);
    m->blockSize = (m->blockSize + BLSIZE - 1)/BLSIZE * BLSIZE;

    m->block = (char *)valloc(  m->blockSize );
    IF (int)(m->block) <= 0 THEN
        perror("Modules0: valloc");
        m->block = NIL;
        return;
    END
    m->SB = (LONGINT)m->block  +  Align8( m->conSize );
    m->PC = (LONGINT)m->block  +  code;
    m->entries = (LONGINT)m->block + entries;
    m->commands = (LONGINT)m->block + commands;
    m->imports = (LONGINT)m->block + imports;
    m->typedescs = (LONGINT)m->block + typedescs;
    m->pointers = (LONGINT)m->block + pointers;
    m->traps = (LONGINT)m->block + traps;
    m->refs = (LONGINT)m->block + refs;
    m->syms = 0;
    m->publics = 0;
    m->privates = 0;
}

/* fixups */

PROCEDURE FixupObj( LONGINT modPC, LONGINT link, LONGINT fixval ) {
    LONGINT instr, nextlink, u, l; 
    LONGINT* code;
    
    code = (LONGINT*)modPC;
    do {
        instr = code[link]; 
        IF instr26(instr) == iBL THEN
            nextlink = -disp26(instr);
            instr = instr26(instr); 
            l = fixval - (modPC + link*4);
            code[link] = instr | l & iBLmask;
        ELSIF opcode(instr) == iCAU THEN
            nextlink = -disp16(instr); instr = instr16(instr); 
            split(fixval, u, l);
            code[link] = instr | u;
            instr = instr16(code[link+1]);
            code[link+1] = instr | l;
        ELSE  assert(FALSE);  
        END
        link = nextlink;
    } while (link != 0);
}

PROCEDURE Fixup( Module mod, LinkEntry* links, LONGINT noflinks,
        	 INTEGER tdadr[], LONGINT linktbl ) {
    LONGINT *data;
    LONGINT i, j, len, basemod, baseadr, nofmth, entno, link, nextlink, instr, fixval;
    LONGINT u;
    INTEGER ent;
    Tag tag, basetag;
    SuperTag supertag;
    LinkEntry entry;
    Module m;
    
    /* fixup link table */
    data = (LONGINT*)(mod->SB + linktbl);
    data[0] = KernelSB;
    FOR i = 0, len = mod->nofImps; i < len; i++  DO
        m = (Module)(*(long*)(mod->imports + 4*i));
        data[i+1] = m->SB;
    END
    /* fixup type descriptors */
    data = (LONGINT*)mod->SB;
    FOR i = 0; i < mod->nofTds; /*nothing!*/  DO
        tag = (Tag)data[tdadr[i]>>2]; j = 0;
        basemod = SHIGH(BASETAG(*tag, 0)); 
        baseadr = SLOW(BASETAG(*tag, 0));
        IF basemod >= 0 THEN
            IF basemod == 0 THEN
                m = mod;
                FOR j = 0; j < mod->nofTds && tdadr[j] != baseadr; j++  DO
                END
                IF j > i THEN 
                    u = tdadr[i]; tdadr[i] = tdadr[j]; tdadr[j] = u;
                    continue;
                END
            ELSE
                m = (Module)(*(LONGINT*)(mod->imports + 4*(basemod-1)));
            END
            basetag = (Tag) *((LONGINT*)(m->SB+baseadr)); j = 0;
            WHILE (BASETAG(*tag, j) = BASETAG(*basetag, j)) != 0 DO j++; END
        END
        BASETAG(*tag, j) = (LONGINT)tag;
        supertag = (SuperTag)(TAG(*tag) & TagMask);
        supertag->ext0 = j++;
        WHILE j < MaxExts  DO
            BASETAG(*tag, j) = 0; j++;
        END
        nofmth = supertag->pad[0]; supertag->pad[0] = 0;
        FOR j = 0; j < nofmth; j++  DO
            IF (entno = METHOD(*tag, j).PC) < 0 THEN
                /* inherited */
                METHOD(*tag, j).PC = METHOD(*basetag, j).PC;
                METHOD(*tag, j).SB = METHOD(*basetag, j).SB;
            ELSE
                /* overwritten */
		ent = ((INTEGER*)mod->entries)[entno];
                METHOD(*tag, j).PC = ent*4 + mod->PC;
                METHOD(*tag, j).SB = mod->SB;
            END
        END
        i++;
    END
    /* fixup code section */
    i = 0;
    WHILE i < noflinks DO
        entry = *links++; 
        IF entry.mod == LinkMTag && entry.ent == CaseETag THEN    
            /* fixup new style case statements */
            link = entry.offset >> 2;
            do {
                nextlink = data[link] >> 18;
                instr = data[link+1] >> 16;    /* size */
                WHILE instr >= 0 THEN
                    data[link+instr] = 
                        (data[link+instr] & 0xffff)*4 + mod->PC;
                    instr--;
                END
                link = nextlink;
            } while (link != 0);
        ELSE
            /* fixup procedure calls or assignments */
            IF entry.mod == SYSMTag THEN
                /* NEW procedures and FP save/restore */
                IF entry.ent == NewRecETag THEN
                    fixval = newRec;
                ELSIF entry.ent == NewSysETag THEN
                    fixval = newSys;
                ELSIF entry.ent == NewArrETag THEN
                    fixval = newArr;
                ELSIF entry.ent < 33 THEN
                    fixval = ((LONGINT)&SaveF) + entry.ent*4;
                ELSIF 32 < entry.ent  &&  entry.ent < 65 THEN
                    fixval = ((LONGINT)&RestF) + (entry.ent-32)*4;
                END
            ELSIF entry.mod == 0 THEN
                /* same module */
                fixval =  mod->PC + ((INTEGER*)mod->entries)[entry.ent]*4;
            ELSE
		m = ((Module*)(mod->imports))[entry.mod-1];
                fixval =  m->PC + ((INTEGER*)m->entries)[entry.ent]*4;
            END
            link = -entry.offset;
            FixupObj( mod->PC, link, fixval );
        END
	i++;
    END
}

PROCEDURE FindTDAdr( Module m, LONGINT fp, LONGINT* adr ) 
{
    ExportPtr  elem;
    LONGINT dsc;
    INTEGER i;
    
    dsc = m->export.dsc; elem = (ExportPtr)dsc;
    i = 0;
    WHILE (i < m->export.nofExp) & (elem->fp != fp)  DO
        i++;
        elem = &((ExportPtr)(dsc))[i];
    END 
    IF i < m->export.nofExp THEN
        elem = (ExportPtr)elem->dsc; 
        *adr = -elem->adr;
    ELSE
        *adr = 0;
    END
}


PROCEDURE LoadScope( File f, 
		ExportPtr scope, INTEGER level, LONGINT adr, 
		ExportPtr Struct[], INTEGER* nofStr ) {
    INTEGER no, no2;
    LONGINT fp, off, arsize;
    ExportPtr dsc, elem;

    ReadInt( f, &scope->nofExp ); no = 0; no2 = 0;
    IF scope->nofExp > 0 THEN 
        arsize = scope->nofExp * sizeof( ExportDesc );
        scope->dsc = (uint)malloc( arsize + 7 ) & ~7;
        dsc = (ExportPtr)scope->dsc;  dsc->adr = adr;
    END
    IF level == EUrecScope THEN
	(*nofStr)++;
        Struct[*nofStr] = scope; 
    END
    ReadNum( f, &fp );
    WHILE fp != EUEnd  DO
        IF fp == EURecord THEN
            ReadNum( f, &off );
            elem = &dsc[no2];
            IF off < 0 THEN
		/* old type */
                elem->nofExp = Struct[-off]->nofExp; 
                elem->dsc    = Struct[-off]->dsc;
            ELSE 
                LoadScope( f, elem, EUrecScope, off, Struct, nofStr );
            END
        ELSE
            elem = &dsc[no];
            IF level == EUobjScope THEN
                ReadNum( f, &elem->adr );
            END
            elem->fp = fp; no2 = no; no++;
        END
        ReadNum( f, &fp );
    END
}

void LoadExpBlock( File ObjFile, Module m ) {
    ExportPtr Struct[1024]; 
    INTEGER nofStr;

    nofStr = 0; 
    LoadScope( ObjFile, &m->export, EUobjScope, 0, Struct, &nofStr );
}

PROCEDURE CheckScope( File ObjFile, 
		Module M, Module m, ExportPtr scope, INTEGER level ) {
    LONGINT fp, link, i, tdadr, dsc; 
    ExportPtr elem, tadr; 
    BOOLEAN tmpErr;
    ModuleName name;

    tmpErr = (level == EUerrScope);
    dsc = scope->dsc; i = 0;
    ReadNum( ObjFile, &fp );
    WHILE fp != EUEnd  DO
        IF fp == EURecord THEN 
            ReadNum( ObjFile, &link );
            elem  =  &((ExportPtr)(dsc))[i];
            IF tmpErr THEN
                 CheckScope( ObjFile, M, m, elem, EUerrScope );
            ELSE
                IF link != 0 THEN
                    tadr = (ExportPtr)elem->dsc; 
                    tdadr = *(LONGINT*)(m->SB - tadr->adr); 
                    *(LONGINT*)(M->SB - link) = tdadr; 
                END
                CheckScope(  ObjFile, M, m, elem, EUrecScope );
            END
        ELSE 
            ReadString( ObjFile, name );
            IF level >= EUobjScope THEN
                tmpErr = FALSE;
                IF level == EUobjScope THEN  ReadNum( ObjFile, &link ); END
                elem = (ExportPtr)dsc;
                i = 0; 
                WHILE (i < scope->nofExp) & (elem->fp != fp)  DO
                    i++; elem = &((ExportPtr)(dsc))[i];
                END
                IF i >= scope->nofExp THEN 
                    IF res == 0 THEN
                        printf( "%s imports incomp. vers. of %s: ",
			M->name, m->name );
                        res = 16;
                    END
                    tmpErr = TRUE;
                    IF name[0] == '@' THEN
                        printf( "\tRECORD incompatible\n" );
                    ELSE
                        printf( "\t%s incompatible\n", name ); 
                    END
                ELSIF (level == EUobjScope) & (link != 0) THEN
                    IF ( link & EUProcFlag ) == 0 THEN
                        *(LONGINT*)(M->SB - link) = m->SB + elem->adr; 
                    ELSE
                        FixupObj( M->PC, link - EUProcFlag, 4*elem->adr + m->PC ); 
                    END
                END
            END
        END
        ReadNum( ObjFile, &fp );
    END
}

PROCEDURE CheckUseBlock( File ObjFile, Module M ) {
    Module mod;
    ModuleName name;

    ReadString( ObjFile, name );
    WHILE (name[0] != 0) & (res == 0) DO
        mod = ThisMod( name );
        IF res == 0 THEN
            CheckScope( ObjFile, M, mod, &mod->export, EUobjScope );
        END
        ReadString( ObjFile, name );
    END
}

static ExportPtr AllocPtrArr( int nofel ) 
{
    LONGINT size, first, last;
    LONGINT* p;
    Tag ptr;
    
    size = tagSize + 12 + 8 + 4*nofel;
    ptr = (Tag)((LONGINT)(calloc( (size+15)/8, 8 )) + 4);
    TAG(*ptr) =  exportDT | 0x2;
    p = (LONGINT*)ptr;
    first = ((LONGINT)ptr) + 20;
    last  = first + 4*nofel - 4;
    *p = last; p++;
    *p = 0; p++;
    *p = first; p++;
    *p = nofel;
    return ((ExportPtr)ptr);
}

PROCEDURE InitGlue( Module m ) {
    /* establish Oberon <-> C interconnection */
    LONGINT* vp;

    vp = (LONGINT*)(m->SB + 20);
    *vp = (LONGINT)&Oberon_dlsym;
}


/* loading an Oberon module */

static Module LoadModule( File ObjFile ) {
    LONGINT BodyPC, i, j;
    Module m, imp;
    LinkEntry links[MaxLinks];
    INTEGER tdadr[MaxRecs];
    Tag tag;
    SuperTag supertag;
    LONGINT size, baseadr;
    INTEGER basemod, nofmth, nofinhmth, nofnewmth, mthno, entno;
    INTEGER nofLinks, lnkTbl, nofptrs, mno;
    ModuleName mname;
    INTEGER *intp;
    LONGINT *lintp;
    LONGINT n;
    SHORTINT modl, entl;


    res = done;
    m = (Module) ((uint)malloc( sizeof( ModuleDesc ) + 7 ) & ~7);
    IF m == NIL THEN
        printf("not enough memory\n");
        res = notEnoughMemory; 
        close(ObjFile->handle);
        return NIL;
    END
    
    /* header block */
    Check( ObjFile, 0xBB ); 
    Check( ObjFile, 0xCF ); 
    
    m->import = AllocPtrArr( DefMaxImport ); 
    m->structs = AllocPtrArr( DefMaxStruct );
    m->reimp = AllocPtrArr( DefMaxReimp );
    m->nofImps = -1; m->refCnt = -1; 
    m->term.PC = 0; m->term.SB = 0;
    m->link = modList; modList = m;
    
    ReadNum( ObjFile, &size ); Skip( ObjFile, size );
    
    ReadLInt( ObjFile, &m->refSize );  ReadInt( ObjFile, &m->nofEntries ); 
    ReadInt(  ObjFile, &m->nofCmds );  ReadInt( ObjFile, &m->nofPtrs ); 
    ReadInt(  ObjFile, &m->nofTds );   ReadInt( ObjFile, &m->nofImps ); 
    ReadInt(  ObjFile, &lnkTbl );      ReadInt( ObjFile, &nofLinks );
    ReadLInt( ObjFile, &m->dataSize ); ReadInt( ObjFile, &m->conSize);
    ReadInt(  ObjFile, &m->codeSize ); ReadInt( ObjFile, &m->nofTraps);
    ReadString( ObjFile, m->name );
    AllocateModule( m );
    IF m->block == NIL THEN
        printf("not enough memory\n");
        res = notEnoughMemory; 
        close( ObjFile->handle );
        return NIL;
    END
printf("loading %s\n", m->name);
    IF strcmp(m->name, "Glue") == 0 THEN  InitGlue( m );  END

    Check( ObjFile, EntryMark );
    intp = (INTEGER*)m->entries; n = 0;
    WHILE n < m->nofEntries DO
	ReadInt( ObjFile, intp++ ); n++;
    END

    Check( ObjFile, CommandMark );
    FOR i = 0; i < m->nofCmds; i++  DO
        ReadString( ObjFile, ((CommandDesc*)m->commands)[i].name );
        ReadInt( ObjFile, &(((CommandDesc*)m->commands)[i].offset) );
    END

    Check( ObjFile, PointerMark );
    lintp = (LONGINT*)m->pointers; n = 0;
    WHILE n < m->nofPtrs DO
	ReadLInt( ObjFile, lintp++ ); n++;
    END

    Check( ObjFile, ImportMark );
    mno = 0;
    WHILE (mno < m->nofImps) & (res == 0)  DO
        ReadString( ObjFile, mname );
        imp = ThisMod( mname );
        IF res == 0 THEN
            imp->refCnt++;
            *(LONGINT*)(m->imports + 4*mno) = (LONGINT)imp;
            mno++;
        END
    END
    
    Check( ObjFile, LinkMark );
    FOR n=0; n < nofLinks; n++ DO
        ReadInt( ObjFile, &links[n].mod );
        ReadInt( ObjFile, &links[n].ent );
        ReadInt( ObjFile, &(links[n].offset) );
    END
    
    Check( ObjFile, DataMark );
    ReadBlock( ObjFile, (CHAR*)m->block, m->conSize );
    
    Check( ObjFile, ExportMark );
    LoadExpBlock( ObjFile, m );
    
    Check( ObjFile, CodeMark );
    lintp = (LONGINT*)m->PC; 
    FOR n = 0; n < m->codeSize; n++ DO
	ReadLInt( ObjFile, lintp++ );
    END
    
    Check( ObjFile, UseMark );
    CheckUseBlock( ObjFile, m );
    
    IF res == done THEN
        Check( ObjFile, TypeMark );
        FOR i = 0; i < m->nofTds; i++  DO
            ReadLInt( ObjFile, &size ); ReadInt( ObjFile, &tdadr[i] ); 
            ReadInt( ObjFile, &basemod ); ReadLInt( ObjFile, &baseadr );
            IF basemod > 0 THEN
                imp = (Module) *(LONGINT*)(m->imports + 4*basemod - 4);
                FindTDAdr( imp, baseadr, &baseadr );
            END
            ReadInt( ObjFile, &nofmth ); ReadInt( ObjFile, &nofinhmth );
            ReadInt( ObjFile, &nofnewmth ); ReadInt( ObjFile, &nofptrs );
            tag = AllocTypeDesc(size, nofmth, nofptrs);
            supertag = (SuperTag)(TAG(*tag) & TagMask);
            ReadString( ObjFile, supertag->typename );
            supertag->Module = (ADDRESS)m;
            BASETAG(*tag, 0) = (basemod << 16)+(baseadr & 0xffff);
            FOR j = 0; j < nofnewmth; j++  DO
                ReadInt( ObjFile, &mthno ); ReadInt( ObjFile, &entno );
                METHOD(*tag, mthno).PC = entno;
            END
            PTROFF(*tag, nofptrs) = -(nofptrs*4 + 4);
	    lintp = (LONGINT*)&PTROFF(*tag, 0); n = 0;
	    WHILE n < nofptrs DO
		ReadLInt( ObjFile, lintp++ ); n++;
	    END
	    ((ADDRESS*)(m->typedescs))[i] = (ADDRESS)tag;
            *(ADDRESS*)(m->SB + tdadr[i]) = (ADDRESS)tag;
        END
        
        Check( ObjFile, TrapMark );
	intp = (INTEGER*)m->traps; n = 0;
	WHILE n < m->nofTraps DO
	    ReadInt( ObjFile, intp++ ); /* offset */
	    ReadInt( ObjFile, intp++ ); /* trapno */
	    n++;
	END
        
        Check(ObjFile, RefMark);
        ReadBlock( ObjFile, (CHAR*)m->refs, m->refSize);
        
        close( ObjFile->handle );
        Fixup( m, links, nofLinks, tdadr, lnkTbl );
        
	if (mprotect(m->block, m->blockSize, PROT_EXEC|PROT_READ|PROT_WRITE) < 0) {
		perror("Modules0: mprotect");
	}
	cacheflush(m->block, m->blockSize, 3);	
	
	m->refCnt = 0;
	BodyPC = m->PC + (*(INTEGER*)(m->entries))*4;
        CallOberon( BodyPC, m->SB );
        res = done;
        return m;
    ELSE
        printf("loading module %s failed\n", m->name);
        WHILE mno > 0  DO
            mno--;
            imp = (Module) *(LONGINT*)(m->imports + 4*mno);
            imp->refCnt--;
        END
        m->refCnt = 0;
        IF modList ==  m THEN
            modList =  m->link;
        ELSE
            imp = modList;
            WHILE imp->link != m)  imp = imp->link;
            imp->link = m->link;
        END
        /* free( (char *)((int)(m->block) - tagSize)); allocated by valloc! */
        close( ObjFile->handle );
	return NIL;
    END
}


/* Oberon high level interface */

static Module ThisMod( CHAR *name ) {
    Module m;
    FileDesc f;

    res = done;
    IF (m = FindModule(name)) == NIL THEN
        strcpy( f.name, name); strcat( f.name, ".Obj" );
        IF FindFile( (File)&f ) == FALSE THEN
            printf("file %s not found\n", f.name);
            res = fileNotFound;
            return NIL;
        ELSE
            /* Oberon .Obj found */
            m = LoadModule( (File)&f );
        END
    END 
    return m;
}

/* initialization and closing */

static Tag CreateTD( Module mod, CHAR* name, LONGINT recsize, 
        LONGINT ptroff[], LONGINT nofptrs, Tag basetag, 
        LONGINT extlev, PROC methods[], LONGINT nofmth ) {
    Tag tag;
    SuperTag supertag;
    LONGINT i;

    tag = AllocTypeDesc( recsize, nofmth, nofptrs );
    for (i = 0; i < nofptrs; i++) PTROFF(*tag, i) = ptroff[i];
    for (i = 0; i < extlev; i++) BASETAG(*tag, i) = BASETAG(*basetag, i);
    BASETAG(*tag, extlev) = (LONGINT)tag;
    for (i = extlev+1; i < MaxExts; i++) BASETAG(*tag, i) = 0;
    for (i = 0; i < nofmth; i++) METHOD(*tag, i) = methods[i];
    supertag = (SuperTag)(TAG(*tag) & TagMask);
    supertag->ext0 = extlev;
    strcpy( supertag->typename, name );
    supertag->Module = (ADDRESS)mod;
    return tag;
}


PROCEDURE Modules0Init() {
    LONGINT i;
    CHAR* path;

    modList =  NIL;
    trace = FALSE;
    KernelSB = 0;

    exportDT = (LONGINT)CreateTD( NIL, "ExportDesc", sizeof( ExportDesc ), NIL, 0, NIL, 0, NIL, 0 );
    res = done;


    /* setup the Oberon path */
    path = getenv( "OBERON" );
    IF path != NIL THEN
        OberonPath = (CHAR*)malloc( strlen( path ) + 1 );
        strcpy( OberonPath, path );
        i = 0;
        WHILE OberonPath[i] != '\0'  DO
            IF OberonPath[i] == ':' THEN  OberonPath[i] = ' ';  END
            i++;
        END
    ELSE
        OberonPath = ". /usr/local/Oberon";
    END
}

extern void LoadKernel( void ) {
    Module m;

    m = ThisMod( "Kernel" );
    IF m != NIL THEN
        KernelSB = m->SB;
    ELSE
        printf( "could not load 'Kernel', terminating\n" ); exit( 3 );
    END
}


int main(int argc, char **argv)
{
	char* p;

	u_argc = argc;
	u_argv = argv;

	printf("Oberon Loader 19.01.2003\n");

	if (geteuid() == 0) {
		suid_root = 1;
		seteuid( getuid() );
	} else {
		suid_root = 0;
	}

	o_debug = 0;
	p = getenv("OBERON_DEBUG");
	if (p != NULL) o_debug = atoi(p);

	Modules0Init(); 
	ThisMod("Glue");
	ThisMod("Console");
	ThisMod("Threads");
	ThisMod("Unix");
	LoadKernel();
	ThisMod("FileDir");
	ThisMod("Files");
	ThisMod("Modules");
	return 0;
}




